import { __awaiter, __generator } from "tslib";
import { isDefined } from "../utils";
export var UpdateFailureReason;
(function (UpdateFailureReason) {
    UpdateFailureReason[UpdateFailureReason["AlreadyUpdated"] = 0] = "AlreadyUpdated";
    UpdateFailureReason[UpdateFailureReason["BadRequest"] = 1] = "BadRequest";
    UpdateFailureReason[UpdateFailureReason["RPCError"] = 2] = "RPCError";
})(UpdateFailureReason || (UpdateFailureReason = {}));
export function isUpdateFailureReason(x) {
    return Number.isInteger(x);
}
/**
 * Base class for all clients to extend.
 */
var BaseAbstractClient = /** @class */ (function () {
    /**
     * Creates a new client.
     * @param cachingMechanism The caching mechanism to use for this client. If not provided, the client will not rely on an external cache.
     */
    function BaseAbstractClient(eventSearchConfig, cachingMechanism) {
        if (eventSearchConfig === void 0) { eventSearchConfig = { fromBlock: 0, maxBlockLookBack: 0 }; }
        this.eventSearchConfig = eventSearchConfig;
        this.cachingMechanism = cachingMechanism;
        this.firstBlockToSearch = 0;
        this.latestBlockSearched = 0;
        this._isUpdated = false;
    }
    Object.defineProperty(BaseAbstractClient.prototype, "isUpdated", {
        /**
         * Indicates whether the client has been updated since it was created.
         * @returns Whether the client has been updated since it was created.
         */
        get: function () {
            return this._isUpdated;
        },
        /**
         * Sets whether the client has been updated since it was created.
         * @param value Whether the client has been updated since it was created.
         * @throws Error if the client has already been updated and the value is false.
         */
        set: function (value) {
            if (this._isUpdated === true && !value) {
                throw new Error("Cannot set isUpdated to false once it is true");
            }
            this._isUpdated = value;
        },
        enumerable: false,
        configurable: true
    });
    /**
     * Validates and updates the stored EventSearchConfig in advance of an update() call.
     * Use isEventSearchConfig() to discriminate the result.
     * @provider Ethers RPC provider instance.
     * @returns An EventSearchConfig instance if valid, otherwise an UpdateFailureReason.
     */
    BaseAbstractClient.prototype.updateSearchConfig = function (provider) {
        return __awaiter(this, void 0, void 0, function () {
            var fromBlock, toBlock, maxBlockLookBack;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        fromBlock = this.firstBlockToSearch;
                        toBlock = this.eventSearchConfig.toBlock;
                        if (!isDefined(toBlock)) return [3 /*break*/, 1];
                        if (fromBlock > toBlock) {
                            throw new Error("Invalid event search config fromBlock (".concat(fromBlock, ") > toBlock (").concat(toBlock, ")"));
                        }
                        return [3 /*break*/, 3];
                    case 1: return [4 /*yield*/, provider.getBlockNumber()];
                    case 2:
                        toBlock = _a.sent();
                        if (toBlock < fromBlock) {
                            return [2 /*return*/, UpdateFailureReason.AlreadyUpdated];
                        }
                        _a.label = 3;
                    case 3:
                        maxBlockLookBack = this.eventSearchConfig.maxBlockLookBack;
                        return [2 /*return*/, { fromBlock: fromBlock, toBlock: toBlock, maxBlockLookBack: maxBlockLookBack }];
                }
            });
        });
    };
    /**
     * Asserts that the client has been updated.
     */
    BaseAbstractClient.prototype.assertUpdated = function () {
        if (!this.isUpdated) {
            throw new Error("Client not updated");
        }
    };
    /**
     * Determines if the client has an external cache.
     * @returns Whether the client has an external cache.
     */
    BaseAbstractClient.prototype.hasCachingMechanism = function () {
        return isDefined(this.cachingMechanism);
    };
    return BaseAbstractClient;
}());
export { BaseAbstractClient };
//# sourceMappingURL=BaseAbstractClient.js.map